// Copyright (C) 2004 MySQL AB
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as published by
// the Free Software Foundation
//
// There are special exceptions to the terms and conditions of the GPL 
// as it is applied to this software. View the full text of the 
// exception in file EXCEPTIONS in the directory of this software 
// distribution.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 

using System;
using System.Text;
using System.Collections;
using MySql.Data.Common;

namespace MySql.Data.MySqlClient
{
	/// <summary>
	/// Summary description for CharSetMap.
	/// </summary>
	internal class CharSetMap
	{
		private static Hashtable	mapping;
		private static string[]		indexMap;

		/// <summary>
		/// Returns the text encoding for a given MySQL character set name
		/// </summary>
		/// <param name="version">Version of the current driver that is making the request</param>
		/// <param name="CharSetName">Name of the character set to get the encoding for</param>
		/// <returns>Encoding object for the given character set name</returns>
		public static Encoding GetEncoding( DBVersion version, string CharSetName ) 
		{
			if (mapping == null )
				InitializeMapping(version);
			try 
			{
				if (! mapping.Contains( CharSetName ))
					throw new MySqlException("Character set '" + CharSetName + "' is not supported");

				object value = mapping[ CharSetName ];
				if (value is string)
					return Encoding.GetEncoding( (string)value );
				return Encoding.GetEncoding( (int)value );
			}
			catch (System.NotSupportedException) 
			{
				return Encoding.GetEncoding(0);
			}
		}

		public static Encoding GetEncoding( DBVersion version, int index )
		{
			if (mapping == null )
				InitializeMapping(version);

			if (index == 0) return null;
			if (index >= indexMap.Length)
				throw new MySqlException("Invalid character set index: " + index);
			return GetEncoding( version, indexMap[index].ToString() );
		}

		/// <summary>
		/// 
		/// </summary>
		private static void InitializeMapping(DBVersion version)
		{
			LoadCharsetMap(version);
			LoadIndexMap();
		}

		private static void LoadIndexMap()
		{
			indexMap = new String[95];

			indexMap[1] = "big5";
			indexMap[2] = "czech";
			indexMap[3] = "dec8";
			indexMap[4] = "dos";
			indexMap[5] = "german1";
			indexMap[6] = "hp8";
			indexMap[7] = "koi8_ru";
			indexMap[8] = "latin1";
			indexMap[9] = "latin2";
			indexMap[10] = "swe7";
			indexMap[11] = "usa7";
			indexMap[12] = "ujis";
			indexMap[13] = "sjis";
			indexMap[14] = "cp1251";
			indexMap[15] = "danish";
			indexMap[16] = "hebrew";
			indexMap[18] = "tis620";
			indexMap[19] = "euc_kr";
			indexMap[20] = "estonia";
			indexMap[21] = "hungarian";
			indexMap[22] = "koi8_ukr";
			indexMap[23] = "win1251ukr";
			indexMap[24] = "gb2312";
			indexMap[25] = "greek";
			indexMap[26] = "win1250";
			indexMap[27] = "croat";
			indexMap[28] = "gbk";
			indexMap[29] = "cp1257";
			indexMap[30] = "latin5";
			indexMap[31] = "latin1_de";
			indexMap[32] = "armscii8";
			indexMap[33] = "utf8";
			indexMap[34] = "win1250ch";
			indexMap[35] = "ucs2";
			indexMap[36] = "cp866";
			indexMap[37] = "keybcs2";
			indexMap[38] = "macce";
			indexMap[39] = "macroman";
			indexMap[40] = "pclatin2";
			indexMap[41] = "latvian";
			indexMap[42] = "latvian1";
			indexMap[43] = "maccebin";
			indexMap[44] = "macceciai";
			indexMap[45] = "maccecias";
			indexMap[46] = "maccecsas";
			indexMap[47] = "latin1bin";
			indexMap[48] = "latin1cias";
			indexMap[49] = "latin1csas";
			indexMap[50] = "cp1251bin";
			indexMap[51] = "cp1251cias";
			indexMap[52] = "cp1251csas";
			indexMap[53] = "macromanbin";
			indexMap[54] = "macromancias";
			indexMap[55] = "macromanciai";
			indexMap[56] = "macromancsas";
			indexMap[57] = "cp1256";
			indexMap[63] = "binary";
			indexMap[64] = "armscii";
			indexMap[65] = "ascii";
			indexMap[66] = "cp1250";
			indexMap[67] = "cp1256";
			indexMap[68] = "cp866";
			indexMap[69] = "dec8";
			indexMap[70] = "greek";
			indexMap[71] = "hebrew";
			indexMap[72] = "hp8";
			indexMap[73] = "keybcs2";
			indexMap[74] = "koi8r";
			indexMap[75] = "koi8ukr";
			indexMap[77] = "latin2";
			indexMap[78] = "latin5";
			indexMap[79] = "latin7";
			indexMap[80] = "cp850";
			indexMap[81] = "cp852";
			indexMap[82] = "swe7";
			indexMap[83] = "utf8";
			indexMap[84] = "big5";
			indexMap[85] = "euckr";
			indexMap[86] = "gb2312";
			indexMap[87] = "gbk";
			indexMap[88] = "sjis";
			indexMap[89] = "tis620";
			indexMap[90] = "ucs2";
			indexMap[91] = "ujis";
			indexMap[92] = "geostd8";
			indexMap[93] = "geostd8";
			indexMap[94] = "latin1";
		}

		private static void LoadCharsetMap(DBVersion version)
		{
			mapping = new Hashtable();

			mapping.Add( "big5", "big5" );		// Traditional Chinese
			mapping.Add( "sjis", "sjis" );		// Shift-JIS
			mapping.Add( "gb2312", "gb2312" );
			if (version.isAtLeast(4,1,2))
				mapping.Add( "latin1", 1252 );	//Latin alphabet #1
			else
				mapping.Add( "latin1", "latin1" );
			mapping.Add( "latin2", "latin2" );
			mapping.Add( "latin3", "latin3" );
			mapping.Add( "latin4", "latin4" );
			mapping.Add( "latin5", "latin5" );
			mapping.Add( "greek", "greek" );
			mapping.Add( "hebrew", "hebrew" );
			mapping.Add( "utf8", "utf-8" );
			mapping.Add( "uc2", "uc-2" );
			mapping.Add( "cp1251cias", 1251 );
			mapping.Add( "cp1251csas", 1251 );


			mapping.Add( "tis620", 874 );

			//TODO:  is this right?
			mapping.Add( "binary", "latin1" );

			// relatively sure about
/*			mapping.Add( "default", 0 );
			mapping.Add( "cp1251", 1251 );		// Russian
			mapping.Add( "win1251", 1251 );
			mapping.Add( "gbk", 936 );			// Simplified Chinese
			mapping.Add( "cp866", 866 );
			mapping.Add( "euc_kr", 949 );

			// maybe, maybe not...
			mapping.Add( "win1250", 1250 );		// Central Eurpoe
			mapping.Add( "win1251ukr", 1251 );
			mapping.Add( "latin1_de", 1252 );	// Latin1 German
			mapping.Add( "german1", 1252 );		// German
			mapping.Add( "danish", 1252 );		// Danish
			mapping.Add( "dos", 437 );			// Dos
			mapping.Add( "pclatin2", 852 );		
			mapping.Add( "win1250ch", 1250 );
			mapping.Add( "cp1257", 1257 );
			mapping.Add( "usa7", 646 );
			mapping.Add( "czech", 912 );
			mapping.Add( "hungarian", 912 );
			mapping.Add( "croat", 912 ); */

			/*			("gb2312", "EUC_CN");
						("ujis", "EUC_JP");
						("latvian", "ISO8859_13");
						("latvian1", "ISO8859_13");
						("estonia", "ISO8859_13");
						("koi8_ru", "KOI8_R");
						("tis620", "TIS620");
						("macroman", "MacRoman");
						("macce", "MacCentralEurope");
			*/

		}
	}
}


