// Copyright (C) 2004 MySQL AB
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as published by
// the Free Software Foundation
//
// There are special exceptions to the terms and conditions of the GPL 
// as it is applied to this software. View the full text of the 
// exception in file EXCEPTIONS in the directory of this software 
// distribution.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 

using System;
using System.Data;
using MySql.Data.MySqlClient;

namespace MySql.Data.Types
{
	/// <summary>
	/// Summary description for MySqlBinary
	/// </summary>
	internal class MySqlBinary : MySqlValue
	{
		private object	mValue;
		private bool	isBinary;
		private ulong	length;

		public MySqlBinary() 
		{
			dbType = DbType.Binary;
			mySqlDbType = MySqlDbType.Blob;
		}

		public MySqlBinary(MySqlDbType type) : this()
		{
			mySqlDbType = type;
		}

		public MySqlBinary(byte[] val, MySqlDbType type) : this(type)
		{
			mValue = val;
		}

		internal string ToString( System.Text.Encoding encoding ) 
		{
			if (mValue is byte[])
				return encoding.GetString( (byte[])mValue, 0, (int)length );
			return mValue.ToString();
		}

		internal override void Serialize(PacketWriter writer, bool binary, object ourValue, int length)
		{
			byte[] buff = (ourValue as System.Byte[]);
			if ( buff == null )
				throw new MySqlException( "Only byte arrays can be serialized by MySqlBinary" );

			if (binary) 
			{
				writer.WriteLength( length );
				writer.Write( buff, 0, length );
			}
			else 
			{
				if ( writer.Version.isAtLeast(4,1,0))
					writer.WriteStringNoNull( "_binary " );

				writer.WriteByte( (byte)'\'');
				EscapeByteArray( buff, length, writer );
				writer.WriteByte((byte)'\'');
			}
		}

		public override byte[] ToBytes()
		{
			return (byte[])mValue;
		}

		public bool IsBinary 
		{
			get { return isBinary; }
			set { isBinary = value; }
		}

		public ulong Length 
		{
			get { return (ulong)(mValue as byte[]).Length; }
			set { length = value; }
		}

		internal override object ValueAsObject
		{
			get { return mValue; }
			set { Value = value; }
		}


		public object Value
		{
			get { return mValue; }
			set 
			{ 
				if (value is byte[] || value is Array)
					isBinary = true;
				else
					isBinary = false;
				mValue = value; 
			} 
		}

		private void EscapeByteArray( byte[] bytes, int length, PacketWriter writer )
		{
			//int theSize = size == 0 ? bytes.Length : size;
			System.IO.MemoryStream ms = (System.IO.MemoryStream)writer.Stream;
			ms.Capacity += (length * 2);

			for (int x=0; x < length; x++)
			{
				byte b = bytes[x];
				if (b == '\0') 
				{
					writer.WriteByte( (byte)'\\' );
					writer.WriteByte( (byte)'0' );
				}
				
				else if (b == '\\' || b == '\'' || b == '\"')
				{
					writer.WriteByte( (byte)'\\' );
					writer.WriteByte( b );
				}
				else
					writer.WriteByte( b );
			}
		}


		internal override DbType DbType
		{
			get
			{
				if (isBinary) return DbType.Binary;
				return DbType.String;
			}
		}


		internal override Type SystemType
		{
			get { return typeof(byte[]); }
		}

		internal override string GetMySqlTypeName()
		{
			switch (mySqlDbType) 
			{
				case MySqlDbType.TinyBlob:		return "TINY_BLOB";
				case MySqlDbType.MediumBlob:	return "MEDIUM_BLOB";
				case MySqlDbType.LongBlob:		return "LONG_BLOB";
				case MySqlDbType.Blob:			
				default:
					return "BLOB";
			}
		}

		internal override void SetData(PacketReader reader, long length)
		{
			if (length == -1)
				length = (long)reader.GetFieldLength();
			this.length = (ulong)length;
			if (! (mValue is byte[]))
				mValue = new byte[ length ];

			byte[] buf = (mValue as byte[]);
			reader.Read( ref buf, 0, length );
		}

		internal override void Skip(PacketReader reader)
		{
			long len = reader.GetFieldLength();
			reader.Skip( len );
		}

	}
}
