// Copyright (C) 2004 MySQL AB
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as published by
// the Free Software Foundation
//
// There are special exceptions to the terms and conditions of the GPL 
// as it is applied to this software. View the full text of the 
// exception in file EXCEPTIONS in the directory of this software 
// distribution.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 

using System;
using System.Data;
using System.IO;
using MySql.Data.MySqlClient;

namespace MySql.Data.Types
{
	/// <summary>
	/// Summary description for MySqlDateTime.
	/// </summary>
	public class MySqlDateTime : MySqlValue, IConvertible
	{
		private DateTime	mValue;
		private bool		isZero;

		internal MySqlDateTime() 
		{
			dbType = DbType.DateTime;
			mySqlDbType = MySqlDbType.Datetime;
		}

		internal MySqlDateTime(MySqlDbType type) : this()
		{
			if (type == MySqlDbType.Date)
				dbType = DbType.Date;
			else
				dbType = DbType.DateTime;
			mySqlDbType = type;
		}

		internal MySqlDateTime(DateTime val, MySqlDbType type) : this(type)
		{
			mValue = val;
		}

		/// <summary>Indicates if this value is zero</summary>
		public bool IsZero 
		{ 
			get { return isZero; }
		}

		private void SerializeText( PacketWriter writer, DateTime value ) 
		{
			string val = String.Empty;

			switch (mySqlDbType) 
			{
				case MySqlDbType.Date:
					val = value.ToString("yyyy-MM-dd"); break;
				case MySqlDbType.Datetime:
					val = value.ToString("yyyy-MM-dd HH:mm:ss"); break;
				case MySqlDbType.Timestamp:
					if ( writer.Version.isAtLeast(4,1,0))
						val = value.ToString("yyyy-MM-dd HH:mm:ss");
					else
						val = value.ToString("yyyyMMddHHmmss");
					break;
			}
			writer.WriteStringNoNull( "'" + val + "'" );
		}


		internal override void Serialize(PacketWriter writer, bool binary, object value, int length)
		{
			if (! (value is DateTime))
				throw new MySqlException( "Only DateTime objects can be serialized by MySqlDateTime" );

			DateTime dtValue = (DateTime)value;
			if (! binary)
			{
				SerializeText( writer, dtValue );
				return;
			}

			if (mySqlDbType == MySqlDbType.Timestamp)
				writer.WriteByte( 11 );
			else
				writer.WriteByte( 7 );

			writer.WriteInteger( dtValue.Year, 2 );
			writer.WriteByte( (byte)dtValue.Month );
			writer.WriteByte( (byte)dtValue.Day );
			if (mySqlDbType == MySqlDbType.Date) 
			{
				writer.WriteByte( 0 );
				writer.WriteByte( 0 );
				writer.WriteByte( 0 );
			}
			else 
			{
				writer.WriteByte( (byte)dtValue.Hour );
				writer.WriteByte( (byte)dtValue.Minute  );
				writer.WriteByte( (byte)dtValue.Second );
			}
			
			if (mySqlDbType == MySqlDbType.Timestamp)
				writer.WriteInteger( dtValue.Millisecond, 4 );
		}

		/// <summary>Returns this value as byte[]</summary>
		public override byte[] ToBytes()
		{
			PacketWriter p = new PacketWriter();
//			Packet p = new Packet(false);

			byte len = 11;
			if (mySqlDbType == MySqlDbType.Date)
				len = 7;
			p.WriteByte( len );

			p.WriteInteger( mValue.Year, 2 );
			p.WriteByte( (byte)mValue.Month );
			p.WriteByte( (byte)mValue.Day );
			if (mySqlDbType == MySqlDbType.Date) 
			{
				p.WriteByte( 0 );
				p.WriteByte( 0 );
				p.WriteByte( 0 );
			}
			else 
			{
				p.WriteByte( (byte)mValue.Hour );
				p.WriteByte( (byte)mValue.Minute  );
				p.WriteByte( (byte)mValue.Second );
			}
			
			if (mySqlDbType == MySqlDbType.Timestamp)
				p.WriteInteger( mValue.Millisecond, 4 );

			byte[] buff = new byte[len+1]; 
			Array.Copy( (p.Stream as MemoryStream).GetBuffer(), 0, buff, 0, len+1 );
			return buff;
		}

		internal override DbType DbType 
		{
			get 
			{ 
				switch (mySqlDbType) 
				{
					case MySqlDbType.Date:			
					case MySqlDbType.Newdate:
						return DbType.Date;
				}
				return DbType.DateTime;
			}
		}

		internal override string GetMySqlTypeName()
		{
			switch (mySqlDbType) 
			{
				case MySqlDbType.Date: return "DATE";
				case MySqlDbType.Newdate: return "NEWDATE";
				case MySqlDbType.Timestamp: return "TIMESTAMP";
			}
			return "DATETIME";
		}

		internal override object ValueAsObject
		{
			get { return mValue; }
			set { mValue = Convert.ToDateTime(value); }
		}

		/// <summary>Returns this value as a DateTime</summary>
		public DateTime Value
		{
			get { return mValue; }
			set { mValue = value; } 
		}

		private void ParseMySql( string s, bool is41 ) 
		{
			isZero = false;
			switch (mySqlDbType) 
			{
				case MySqlDbType.Date:
					if (s == "0000-00-00") isZero = true;
					else mValue = DateTime.ParseExact( s, "yyyy-MM-dd", System.Globalization.DateTimeFormatInfo.InvariantInfo );
					break;

				case MySqlDbType.Datetime:
					if (s == "0000-00-00 00:00:00") isZero = true;
					else mValue = DateTime.ParseExact( s, "yyyy-MM-dd HH:mm:ss", System.Globalization.DateTimeFormatInfo.InvariantInfo );
					break;

				case MySqlDbType.Timestamp:
					// MySql 4.1.0 and later use DateTime format for timestamp
					if (is41)
					{
						if (s == "0000-00-00 00:00:00") isZero = true;
						else mValue = DateTime.ParseExact( s, "yyyy-MM-dd HH:mm:ss", System.Globalization.DateTimeFormatInfo.InvariantInfo );
						return;
					}

					string pattern;
					switch (s.Length) 
					{
						case 2:  pattern = "yy"; break;
						case 4:  pattern = "yyMM"; break;
						case 6:  pattern = "yyMMdd"; break;
						case 8:  pattern = "yyyyMMdd"; break;
						case 10: pattern = "yyMMddHHmm"; break;
						case 12: pattern = "yyMMddHHmmss"; break;
						case 14: 
						default: pattern = "yyyyMMddHHmmss"; break;
					}

					if (s.Replace("0","").Length == 0) isZero = true;
					else
						mValue = DateTime.ParseExact( s, pattern, System.Globalization.DateTimeFormatInfo.InvariantInfo );
					break;
			}
		}

		internal override Type SystemType
		{
			get { return typeof(DateTime); }
		}

		internal override void SetData(PacketReader reader, long length)
		{
			if (length >= 0) 
			{
				string value = reader.ReadString( length );
				ParseMySql( value, reader.Version.isAtLeast(4,1,0) );
				return;
			}

			long bufLength = reader.ReadByte();

			if (bufLength == 0) 
				IsNull = true;
			else if (bufLength == 4)
				mValue = new DateTime( reader.ReadInteger(2), reader.ReadByte(), 
					reader.ReadByte() );
			else if (bufLength == 7)
				mValue = new DateTime( reader.ReadInteger(2), reader.ReadByte(), 
					reader.ReadByte(), reader.ReadByte(), reader.ReadByte(),
					reader.ReadByte() );
			else if (bufLength == 11)
				mValue = new DateTime( reader.ReadInteger(2), reader.ReadByte(), 
					reader.ReadByte(), reader.ReadByte(), reader.ReadByte(),
					reader.ReadByte(), reader.ReadInteger(4) / 1000000 );
		}

		internal override void Skip(PacketReader reader)
		{
			long len = reader.ReadByte();
			reader.Skip( len );
		}

		/// <summary>Returns a MySQL specific string representation of this value</summary>
		public override string ToString()
		{
			if (mySqlDbType == MySqlDbType.Date)
				return IsZero ? "0000-00-00" : Value.ToString("yyyy-MM-dd");
			return IsZero ? "0000-00-00 00:00:00" : Value.ToString( "yyyy-MM-dd HH:mm:ss" );
		}

		/// <summary></summary>
		/// <param name="val"></param>
		/// <returns></returns>
		public static explicit operator DateTime( MySqlDateTime val ) 
		{
			if (val.IsZero) return DateTime.MinValue;
			return val.Value;
		}

		#region IConvertible Members

		ulong IConvertible.ToUInt64(IFormatProvider provider)
		{
			return Convert.ToUInt64( mValue );
		}

		sbyte IConvertible.ToSByte(IFormatProvider provider)
		{
			return Convert.ToSByte( mValue );
		}

		double IConvertible.ToDouble(IFormatProvider provider)
		{
			return Convert.ToDouble( mValue );
		}

		DateTime IConvertible.ToDateTime(IFormatProvider provider)
		{
			return mValue;
		}

		float IConvertible.ToSingle(IFormatProvider provider)
		{
			return Convert.ToSingle( mValue );
		}

		bool IConvertible.ToBoolean(IFormatProvider provider)
		{
			return Convert.ToBoolean( mValue );
		}

		int IConvertible.ToInt32(IFormatProvider provider)
		{
			return Convert.ToInt32( mValue );
		}

		ushort IConvertible.ToUInt16(IFormatProvider provider)
		{
			return Convert.ToUInt16( mValue );
		}

		short IConvertible.ToInt16(IFormatProvider provider)
		{
			return Convert.ToInt16( mValue );
		}

		string IConvertible.ToString(IFormatProvider provider)
		{
			return ToString();
		}

		byte IConvertible.ToByte(IFormatProvider provider)
		{
			return Convert.ToByte( mValue );
		}

		char IConvertible.ToChar(IFormatProvider provider)
		{
			return Convert.ToChar( mValue );
		}

		long IConvertible.ToInt64(IFormatProvider provider)
		{
			return Convert.ToInt64( mValue );
		}

		TypeCode IConvertible.GetTypeCode()
		{
			return TypeCode.Object;
		}

		decimal IConvertible.ToDecimal(IFormatProvider provider)
		{
			return Convert.ToDecimal( mValue );
		}

		object IConvertible.ToType(Type conversionType, IFormatProvider provider)
		{
			return null;
		}

		uint IConvertible.ToUInt32(IFormatProvider provider)
		{
			return Convert.ToUInt32( mValue );
		}

		#endregion
	}
}
