// Copyright (C) 2004 MySQL AB
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as published by
// the Free Software Foundation
//
// There are special exceptions to the terms and conditions of the GPL 
// as it is applied to this software. View the full text of the 
// exception in file EXCEPTIONS in the directory of this software 
// distribution.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 

using System;
using System.Data;
using System.IO;
using System.Globalization;
using System.Threading;
using NUnit.Framework;

namespace MySql.Data.MySqlClient.Tests
{
	[TestFixture()]
	public class LanguageTests : BaseTest
	{
		[TestFixtureSetUp]
		public void FixtureSetup()
		{
			Open();
		}

		[TestFixtureTearDown]
		public void FixtureTeardown()
		{
			Close();
		}

		[Test]
		[Explicit]
		public void Unicode()
		{
			MySqlCommand cmd = new MySqlCommand(
				"DROP TABLE IF EXISTS test; " +
				"CREATE TABLE test (u1 varchar(255) CHARACTER SET utf8, u2 varchar(255) CHARACTER SET ucs2)", conn);
			cmd.ExecuteNonQuery();

			String teststring = "\u0120\u0121\u0125\u0411";

			cmd.CommandText = "INSERT INTO test VALUES(?p1, ?p2)";
			cmd.Parameters.Add( "?p1", teststring );
			cmd.Parameters.Add( "?p2", teststring );
			cmd.ExecuteNonQuery();

			cmd.CommandText = "SELECT * FROM test";
			MySqlDataReader reader = cmd.ExecuteReader();
			reader.Read();
			string s1 = reader.GetString(0);
			string s2 = reader.GetString(1);
			reader.Close();

			Assert.AreEqual( teststring, s1 );
			Assert.AreEqual( teststring, s2 );
		}

		[Test()]
		public void UTF8() 
		{
			if (! Is40 && !Is41 && !Is50) return;

			execSQL("DROP TABLE IF EXISTS Test");
			execSQL("CREATE TABLE Test (id int, name VARCHAR(200) CHAR SET utf8)");
			
			execSQL( "INSERT INTO Test VALUES(1, 'ЁЄЉҖҚ')" );		// Russian
			execSQL( "INSERT INTO Test VALUES(2, '兣冘凥凷冋')" );	// simplified Chinese
			execSQL( "INSERT INTO Test VALUES(3, '困巫忘否役')" );	// traditional Chinese
			execSQL( "INSERT INTO Test VALUES(4, '涯割晦叶角')" );	// Japanese
			execSQL( "INSERT INTO Test VALUES(5, 'ברחפע')" );		// Hebrew
			execSQL( "INSERT INTO Test VALUES(6, 'ψόβΩΞ')" );		// Greek
			execSQL( "INSERT INTO Test VALUES(7, 'þðüçöÝÞÐÜÇÖ')" );	// Turkish

			MySqlCommand cmd = new MySqlCommand("SELECT * FROM Test", conn);
			MySqlDataReader reader = null;
			try 
			{
				reader = cmd.ExecuteReader();
				reader.Read();
				Assert.AreEqual( "ЁЄЉҖҚ", reader.GetString(1) );
				reader.Read();
				Assert.AreEqual( "兣冘凥凷冋", reader.GetString(1) );
				reader.Read();
				Assert.AreEqual( "困巫忘否役", reader.GetString(1) );
				reader.Read();
				Assert.AreEqual( "涯割晦叶角", reader.GetString(1) );
				reader.Read();
				Assert.AreEqual( "ברחפע", reader.GetString(1) );
				reader.Read();
				Assert.AreEqual( "ψόβΩΞ", reader.GetString(1) );
				reader.Read();
				Assert.AreEqual( "þðüçöÝÞÐÜÇÖ", reader.GetString(1) );
			}
			catch (Exception ex) 
			{
				Assert.Fail( ex.Message );
			}
			finally 
			{
				if (reader != null) reader.Close();
			}
		}

		[Test()]
		public void Chinese() 
		{
			if (! Is40 && !Is41 && !Is50) return;

			execSQL("DROP TABLE IF EXISTS Test");
			execSQL("CREATE TABLE Test (id int, name VARCHAR(200) CHAR SET big5, name2 VARCHAR(200) CHAR SET gb2312)");

			System.Text.Encoding e = System.Text.Encoding.GetEncoding("gb2312");

			execSQL( "INSERT INTO Test VALUES(1, '困巫忘否役', '涝搞谷侪魍' )" );	

			MySqlCommand cmd = new MySqlCommand("SELECT * FROM Test", conn);
			MySqlDataReader reader = null;
			try 
			{
				reader = cmd.ExecuteReader();
				reader.Read();
				Assert.AreEqual( "困巫忘否役", reader.GetString(1) );
				Assert.AreEqual( "涝搞谷侪魍", reader.GetString(2) );
			}
			catch (Exception ex) 
			{
				Assert.Fail( ex.Message );
			}
			finally 
			{
				if (reader != null) reader.Close();
			}
		}

		[Test()]
		public void Turkish() 
		{
			if (! Is40 && !Is41 && !Is50) return;

			execSQL("DROP TABLE IF EXISTS Test");
			execSQL("CREATE TABLE Test (id int, name VARCHAR(200) CHAR SET latin5 )");
			
			execSQL( "INSERT INTO Test VALUES(1, 'þðüçöÝÞÐÜÇÖ')" );	

			MySqlCommand cmd = new MySqlCommand("SELECT * FROM Test", conn);
			MySqlDataReader reader = null;
			try 
			{
				reader = cmd.ExecuteReader();
				reader.Read();
				Assert.AreEqual( "þðüçöÝÞÐÜÇÖ", reader.GetString(1) );
			}
			catch (Exception ex) 
			{
				Assert.Fail( ex.Message );
			}
			finally 
			{
				if (reader != null) reader.Close();
			}
		}

		[Test()]
		public void Russian() 
		{
			if (! Is40 && !Is41 && !Is50) return;

			execSQL("DROP TABLE IF EXISTS Test");
			execSQL("CREATE TABLE Test (id int, name VARCHAR(200) CHAR SET cp1251)");
			
			execSQL( "INSERT INTO Test VALUES(1, 'щьеи')" );		// Russian

			MySqlCommand cmd = new MySqlCommand("SELECT * FROM Test", conn);
			MySqlDataReader reader = null;
			try 
			{
				reader = cmd.ExecuteReader();
				reader.Read();
				Assert.AreEqual( "щьеи", reader.GetString(1) );
			}
			catch (Exception ex) 
			{
				Assert.Fail( ex.Message );
			}
			finally 
			{
				if (reader != null) reader.Close();
			}
		}

	}
}
